import numpy as np

def dist_calc(surf, cortex, src):
    """
    Calculate exact geodesic distance along cortical surface from set of source nodes.
    """
    import gdist

    vertices, triangles = surf_keep_cortex(surf, cortex)
    data = gdist.compute_gdist(vertices, triangles, source_indices=src)
    dist = recort(data, surf, cortex)
    del data

    return dist


def zone_calc(surf, cortex, src):
    """
    Calculate closest nodes to each source node using exact geodesic distance along the cortical surface.
    """
    import gdist

    vertices, triangles = surf_keep_cortex(surf, cortex)

    dist_vals = np.zeros((len(src),len(vertices)))
    for x in range(len(src)):
        dist_vals[x,:] = gdist.compute_gdist(vertices, triangles, 
                            source_indices=np.array([src[x]], dtype=np.int32))

    data = np.argsort(dist_vals, axis=0)[-1,:] + 1
    zone = recort(data, surf, cortex)
    del data
    
    return zone



def viz(surf, data):
    """
    Visualize results on cortical surace using mayavi.
    """
    vertices = np.array(surf[0], dtype=np.float64)
    triangles = np.array(surf[1], dtype=np.int32)
    
    from mayavi import mlab
    mlab.triangular_mesh(
        vertices[:, 0], vertices[:, 1], vertices[:, 2], triangles, scalars=data)




def find_idx_match(simple_vertices, complex_vertices):
    '''
    Thanks to juhuntenburg. 
    Functions taken from https://github.com/juhuntenburg/brainsurfacescripts

    Finds those points on the complex mesh that correspoind best to the
    simple mesh while forcing a one-to-one mapping
    '''
    # make array for writing in final voronoi seed indices
    voronoi_seed_idx = np.zeros((simple_vertices.shape[0],), dtype='int64')-1
    missing = np.where(voronoi_seed_idx == -1)[0].shape[0]
    mapping_single = np.zeros_like(voronoi_seed_idx)

    neighbours = 0
    col = 0

    while missing != 0:

        neighbours += 100
        # find nearest neighbours
        inaccuracy, mapping = spatial.KDTree(
            complex_vertices).query(simple_vertices, k=neighbours)
        # go through columns of nearest neighbours until unique mapping is
        # achieved, if not before end of neighbours, extend number of
        # neighbours
        while col < neighbours:
            # find all missing voronoi seed indices
            missing_idx = np.where(voronoi_seed_idx == -1)[0]
            missing = missing_idx.shape[0]
            if missing == 0:
                break
            else:
                # for missing entries fill in next neighbour
                mapping_single[missing_idx] = np.copy(
                    mapping[missing_idx, col])
                # find unique values in mapping_single
                unique, double_idx = np.unique(
                    mapping_single, return_inverse=True)
                # empty voronoi seed index
                voronoi_seed_idx = np.zeros(
                    (simple_vertices.shape[0],), dtype='int64')-1
                # fill voronoi seed idx with unique values
                for u in range(unique.shape[0]):
                    # find the indices of this value in mapping
                    entries = np.where(double_idx == u)[0]
                    # set the first entry to the value
                    voronoi_seed_idx[entries[0]] = unique[u]
                # go to next column
                col += 1

    return voronoi_seed_idx, inaccuracy


def competetive_fast_marching(vertices, graph, seeds):
    '''
    Label all vertices on highres mesh to the closest seed vertex
    using a balanced binary search tree
    '''
    # make a labelling container to be filled with the search tree
    # first column are the vertex indices of the complex mesh
    # second column are the labels from the simple mesh
    # (-1 for all but the corresponding points for now)
    labels = np.zeros((vertices.shape[0], 2), dtype='int64')-1
    labels[:, 0] = range(vertices.shape[0])
    for i in range(seeds.shape[0]):
        labels[seeds[i]][1] = i
    # initiate AVLTree for binary search
    tree = FastAVLTree()
    # organisation of the tree will be
    # key: edge length; value: tuple of vertices (source, target)
    # add all neighbours of the voronoi seeds
    for v in seeds:
        add_neighbours(v, 0, graph, labels, tree)
    # Competetive fast marching starting from voronoi seeds
    while tree.count > 0:

        # pdb.set_trace()
        # pop the item with minimum edge length
        min_item = tree.pop_min()
        length = min_item[0]
        source = min_item[1][0]
        target = min_item[1][1]
        # if target no label yet (but source does!), assign label of source
        if labels[target][1] == -1:
            if labels[source][1] == -1:
                sys.exit('Source has no label, something went wrong!')
            else:
                # assign label of source to target
                labels[target][1] = labels[source][1]

        # test if labelling is complete
        if any(labels[:, 1] == -1):
            # if not, add neighbours of target to tree
            add_neighbours(target, length, graph, labels, tree)
        else:
            break

        # print 'tree '+str(tree.count)
        # print 'labels '+str(np.where(labels[:,1]==-1)[0].shape[0])

    return labels


def sample_simple(highres_data, labels):
    '''
    Computes the mean of data from highres mesh that is assigned to the same
    label (typical simple mesh vertices).
    '''
    # create new empty lowres data array
    lowres_data = np.empty((int(labels.max()+1), highres_data.shape[1]))
    # find all vertices on highres and mean
    for l in range(int(labels.max())):
        patch = np.where(labels == l)[0]
        patch_data = highres_data[patch]
        patch_mean = np.mean(patch_data, axis=0)
        lowres_data[l] = patch_mean

    return lowres_data



import nibabel as nib
import numpy as np
import os


def load_freesurfer_label(annot_input, label_name):
	"""
	Get source node list for freesurfer label.
	"""

	annot = nib.freesurfer.read_annot(annot_input)
	label_text = label_name
	label_ind = annot[2].index(label_text)
	labels = np.where(np.in1d(annot[0], label_ind))
	labels = np.where(np.in1d(cort, labels))[0]
	src = np.array(labels, dtype=np.int32)

	return src


def get_freesurfer_label(annot_input, label_name):
	"""
	Print freesurfer label names.
	"""
    annot = nib.freesurfer.read_annot(annot_input)
    print annot[2]




from distutils.core import setup
setup(
  name = 'surfdist',
  packages = ['surfdist'],
  install_requires = ['numpy',                                              
                      'mayavi',                        
                      'gdist', 
                      'nibabel',
                      'os',
                      'scipy'],
  version = '0.13.1',
  description = 'For calculating exact geodesic distances on cortical surface meshes',
  author = 'Daniel Margulies',
  author_email = 'margulies@cbs.mpg.de',
  url = 'https://github.com/margulies/surfdist',
  download_url = 'https://github.com/margulies/surfdist/tarball/0.13.1',
  keywords = ['geodesic', 'distance', 'brain', 'cortex'],
  classifiers = [],
)



import numpy as np


def surf_keep_cortex(surf, cortex):
    """
    Remove medial wall from cortical surface to ensure that shortest paths are only calculated through the cortex.
    """
    vertices = np.array(surf[0][cortex], dtype=np.float64)

    keep = np.zeros(len(surf[1]))
    for i in [0, 1, 2]:
        keep += np.array([item in cortex for item in surf[1][:, i]])
    ind = np.where(keep == 3)
    triangles = np.array(surf[1][ind], dtype=np.int32)
    for c, i in enumerate(cortex):
        triangles[np.where(triangles == i)] = c

    return vertices, triangles


def recort(input_data, surf, cortex):
    """
    Return data values to space of full cortex (including medial wall), with medial wall equal to zero.
    """
    data = np.zeros(len(surf[0]))
    data[cortex] = input_data
    return data


def find_node_match(simple_vertices, complex_vertices):
    '''
    Thanks to juhuntenburg. 
    Functions taken from https://github.com/juhuntenburg/brainsurfacescripts

    Finds those points on the complex mesh that correspoind best to the
    simple mesh while forcing a one-to-one mapping.
    '''

    import scipy.spatial

    # make array for writing in final voronoi seed indices
    voronoi_seed_idx = np.zeros((simple_vertices.shape[0],), dtype='int64')-1
    missing = np.where(voronoi_seed_idx == -1)[0].shape[0]
    mapping_single = np.zeros_like(voronoi_seed_idx)

    neighbours = 0
    col = 0

    while missing != 0:

        neighbours += 100
        # find nearest neighbours
        inaccuracy, mapping = scipy.spatial.KDTree(
            complex_vertices).query(simple_vertices, k=neighbours)
        # go through columns of nearest neighbours until unique mapping is
        # achieved, if not before end of neighbours, extend number of
        # neighbours
        while col < neighbours:
            # find all missing voronoi seed indices
            missing_idx = np.where(voronoi_seed_idx == -1)[0]
            missing = missing_idx.shape[0]
            if missing == 0:
                break
            else:
                # for missing entries fill in next neighbour
                mapping_single[missing_idx] = np.copy(
                    mapping[missing_idx, col])
                # find unique values in mapping_single
                unique, double_idx = np.unique(
                    mapping_single, return_inverse=True)
                # empty voronoi seed index
                voronoi_seed_idx = np.zeros(
                    (simple_vertices.shape[0],), dtype='int64')-1
                # fill voronoi seed idx with unique values
                for u in range(unique.shape[0]):
                    # find the indices of this value in mapping
                    entries = np.where(double_idx == u)[0]
                    # set the first entry to the value
                    voronoi_seed_idx[entries[0]] = unique[u]
                # go to next column
                col += 1

    return voronoi_seed_idx, inaccuracy
